  (* Code generation: translate takes a semantically checked AST and
produces LLVM IR

LLVM tutorial: Make sure to read the OCaml version of the tutorial

http://llvm.org/docs/tutorial/index.html

Detailed documentation on the OCaml LLVM library:

http://llvm.moe/
http://llvm.moe/ocaml/

*)

module L = Llvm
module A = Ast

open Sast
module StringMap = Map.Make(String)

let translate (globals, functions) =
  let context     = L.global_context () in
  let the_module  = L.create_module context "YAMML" in 
  let i32_t       = L.i32_type  context in
  let i64_t       = L.i64_type  context in
  let float_t     = L.double_type context in
  let void_t      = L.void_type   context in
  let pointer_t   = L.pointer_type  in
  let i1_t        = L.i1_type     context
  and i8_t        = L.i8_type   context in
  let matrix_t    = L.named_struct_type context "matrix_t" in 
    L.struct_set_body matrix_t [|L.pointer_type float_t; i32_t; i32_t|] false;

  let ltype_of_typ = function
      A.Int -> i32_t
    | A.Bool  -> i1_t
    | A.Char -> i8_t  
    | A.Float -> float_t 
    | A.Void  -> void_t
    | A.String  -> pointer_t i8_t
    | A.Matrix -> matrix_t 
    | A.Rangety -> raise (Failure "internal error: Rangety not an LLVM type") 
  in

  let define_matrix_lit_global(f_array,(r,c)) =
    let lf_array = Array.map (fun f -> L.const_float float_t f) f_array in
    let mat_arr = L.const_array float_t lf_array in
    let mat_def = L.define_global ".mat_array" mat_arr the_module in
    let mat = L.const_gep mat_def [|L.const_int i64_t 0; L.const_int i64_t 0|] in
    L.const_named_struct matrix_t [|mat; L.const_int i32_t r; L.const_int i32_t c|]
  in

  (* Declare each global variable; remember its value in a map *)
  let global_vars =
    let global_var m b =
      match b with 
          SDeclInit((t,n),e) -> 
            let rec const_expr ((_, e) : sexpr) = match e with
              SLiteral i -> L.const_int (ltype_of_typ t) i
            | SBoolLit b  -> L.const_int (ltype_of_typ t) (if b then 1 else 0)
            | SCharLit c -> L.const_int (ltype_of_typ t) (Char.code c)
            | SStringLit s -> let alloc = L.define_global ".str" (L.const_stringz context s) the_module in
                              L.const_in_bounds_gep alloc [|L.const_int i64_t 0; L.const_int i64_t 0|]
            | SMatrixLit (f_array, (r,c)) -> define_matrix_lit_global (f_array, (r, c))
            | SFloatLit l -> L.const_float_of_string float_t l
            | SCastExpr (ct, ((et, _) as e)) ->
              if et = ct then const_expr e else let op = (match et,ct with
                A.Int,A.Float
              | A.Char,A.Float
              | A.Bool,A.Float  -> L.const_sitofp
              | A.Float,A.Int
              | A.Float,A.Char
              | A.Float,A.Bool -> L.const_fptosi
              | A.Int,A.Char
              | A.Int,A.Bool
              | A.Char,A.Bool -> L.const_trunc
              | A.Bool, A.Char
              | A.Bool, A.Int
              | A.Char, A.Int -> L.const_sext
              | _,_ -> raise (Failure ("internal error: semant should have rejected cast " ^ A.string_of_typ et ^ " -> " ^ A.string_of_typ ct)))
              in op (const_expr e) (ltype_of_typ ct)
            | SNoexpr -> raise (Failure "empty global initializer")
            | SBinop ((A.Float,_ ) as e1, op, e2) -> 
              let e1' =  match e1 with 
                (A.Float, _) -> const_expr e1
              | _ -> L.const_sitofp (const_expr e1) float_t
              and e2' = match e2 with
                (A.Float,_) -> const_expr e2
              | _ -> L.const_sitofp (const_expr e2) float_t in
              (match op with 
                A.Add     -> L.const_fadd
              | A.Sub     -> L.const_fsub
              | A.Mult    -> L.const_fmul
              | A.Div     -> L.const_fdiv 
              | A.Equal   -> L.const_fcmp L.Fcmp.Oeq
              | A.Neq     -> L.const_fcmp L.Fcmp.One
              | A.Less    -> L.const_fcmp L.Fcmp.Olt
              | A.Leq     -> L.const_fcmp L.Fcmp.Ole
              | A.Greater -> L.const_fcmp L.Fcmp.Ogt
              | A.Geq     -> L.const_fcmp L.Fcmp.Oge
              | A.And | A.Or -> raise (Failure "internal error: semant should have rejected and/or on float")
              | A.Elemult | A.Elediv -> raise (Failure "internal error: semant should have rejected Elemult/Elediv on float")
              ) e1' e2' 
            | SBinop (e1, op, e2) ->
              let e1' = const_expr e1
              and e2' = const_expr e2 in
              (match op with
                A.Add     -> L.const_add
              | A.Sub     -> L.const_sub
              | A.Mult    -> L.const_mul
              | A.Div     -> L.const_sdiv
              | A.And     -> L.const_and
              | A.Or      -> L.const_or
              | A.Equal   -> L.const_icmp L.Icmp.Eq
              | A.Neq     -> L.const_icmp L.Icmp.Ne
              | A.Less    -> L.const_icmp L.Icmp.Slt
              | A.Leq     -> L.const_icmp L.Icmp.Sle
              | A.Greater -> L.const_icmp L.Icmp.Sgt
              | A.Geq     -> L.const_icmp L.Icmp.Sge
              | A.Elemult | A.Elediv -> raise (Failure "internal error: semant should have rejected Elemult/Elediv on int")
              ) e1' e2' 
            | SUnop(op, ((t, _) as e)) ->
              let e' = const_expr e in
              (match op with
                A.Neg when t = A.Float -> L.const_fneg 
              | A.Neg                  -> L.const_neg
              | A.Not                  -> L.const_not) e'
            | SMatAccess(_,_,_)
            | SMatSlice(_,_,_)
            | SStringAccess(_,_)
            | SId(_)
            | SAssign(_,_)
            | SCall(_,_)
            | SRange(_,_) -> raise (Failure "non-constant global initializer")
          in StringMap.add n (L.define_global n (const_expr e) the_module) m
      | SDecl(t,n) -> 
          let init = L.const_int (ltype_of_typ t) 0
          in StringMap.add n (L.define_global n init the_module) m
    in 
  List.fold_left global_var StringMap.empty globals in

let printf_t : L.lltype = 
    L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
let printf_func = L.declare_function "printf" printf_t the_module in

let ssum_t : L.lltype = 
  L.function_type float_t [| L.pointer_type float_t; i32_t; i32_t|] in
let ssum_func : L.llvalue = L.declare_function "ssum" ssum_t the_module in

let smean_t : L.lltype = 
  L.function_type float_t [| L.pointer_type float_t; i32_t; i32_t|] in
let smean_func : L.llvalue = L.declare_function "smean" smean_t the_module in

let strans_t : L.lltype = L.function_type void_t [| L.pointer_type float_t; L.pointer_type float_t; i32_t; i32_t|] in
let strans_func : L.llvalue = L.declare_function "strans" strans_t the_module in

let sfilter2D_t : L.lltype = L.function_type void_t [| L.pointer_type matrix_t; L.pointer_type matrix_t; i32_t; L.pointer_type matrix_t|] in
let sfilter2D_func : L.llvalue = L.declare_function "sfilter2D" sfilter2D_t the_module in

let simread_t : L.lltype = L.function_type void_t [|pointer_t i8_t; pointer_t matrix_t|] in
let simread_func : L.llvalue = L.declare_function "simread" simread_t the_module in

let simwrite_t : L.lltype = L.function_type i1_t [|pointer_t i8_t; L.pointer_type matrix_t|] in
let simwrite_func : L.llvalue = L.declare_function "simwrite" simwrite_t the_module in

let sconsmul_t : L.lltype = L.function_type void_t [| L.pointer_type float_t; L.pointer_type float_t; i32_t; i32_t; float_t|] in
let sconsmul_func : L.llvalue = L.declare_function "sconsmul" sconsmul_t the_module in

let sconsdiv_t : L.lltype = L.function_type void_t [| L.pointer_type float_t; L.pointer_type float_t; i32_t; i32_t; float_t|] in
let sconsdiv_func : L.llvalue = L.declare_function "sconsdiv" sconsdiv_t the_module in

let smatmul_t : L.lltype = L.function_type void_t [| L.pointer_type float_t; L.pointer_type float_t; L.pointer_type float_t; i32_t; i32_t; i32_t; i32_t|] in
let smatmul_func : L.llvalue = L.declare_function "smatmul" smatmul_t the_module in

let selemul_t : L.lltype = L.function_type void_t [| L.pointer_type float_t; L.pointer_type float_t; L.pointer_type float_t; i32_t; i32_t; i32_t; i32_t|] in
let selemul_func : L.llvalue = L.declare_function "selemul" selemul_t the_module in
let selediv_t : L.lltype = L.function_type void_t [| L.pointer_type float_t; L.pointer_type float_t; L.pointer_type float_t; i32_t; i32_t; i32_t; i32_t|] in
let selediv_func : L.llvalue = L.declare_function "selediv" selediv_t the_module in
let sprintm_t : L.lltype = L.function_type i32_t [| L.pointer_type float_t; i32_t; i32_t|] in
let sprintm_func : L.llvalue = L.declare_function "sprintm" sprintm_t the_module in

let smatslice_t : L.lltype = L.function_type void_t [| L.pointer_type matrix_t; i32_t; i32_t; i32_t; i32_t; i32_t; i32_t; L.pointer_type matrix_t|] in
let smatslice_func : L.llvalue = L.declare_function "smatslice" smatslice_t the_module in

let sinvert_t : L.lltype = L.function_type void_t [| L.pointer_type matrix_t; L.pointer_type matrix_t|] in
let sinvert_func : L.llvalue = L.declare_function "sinvert" sinvert_t the_module in

let function_decls =
  let function_decl m fdecl =
    let name = fdecl.sfname
    and formal_types =
        Array.of_list (List.map (fun (t,_) -> ltype_of_typ t) fdecl.sformals)
          in let ftype = L.function_type (ltype_of_typ fdecl.styp) formal_types in
          StringMap.add name (L.define_function name ftype the_module, fdecl) m in
      List.fold_left function_decl StringMap.empty functions in

  let build_function_body fdecl = 
    let (the_function, _) = StringMap.find fdecl.sfname function_decls in
    let builder = L.builder_at_end context (L.entry_block the_function) in

    let string_format_str = L.build_global_stringptr "%s\n" "fmt" builder in
    let int_format_str = L.build_global_stringptr "%d\n" "fmt" builder in
    let float_format_str = L.build_global_stringptr "%g\n" "fmt" builder in
    let char_format_str = L.build_global_stringptr "%c\n" "fmt" builder in
    

    (* Return the value for a variable or formal argument.
        Check local names first, then global names *)
    let lookup map n = try StringMap.find n map
                    with Not_found -> try StringMap.find n global_vars
                    with Not_found -> raise (Failure ("Not defined: " ^ n))
    in

    let build_matrix_lit(f_array,(r,c)) builder =
      let mat = L.build_array_malloc float_t (L.const_int i32_t (r*c)) "system_mat" builder in
      (for x = 0 to (r*c-1) do
        let element_ptr = L.build_gep mat [|(L.const_int i32_t x)|] "element_ptr" builder in 
        ignore(L.build_store (L.const_float float_t f_array.(x)) element_ptr builder)
      done);
      let m = L.build_alloca matrix_t "m" builder in
      let m_mat = L.build_struct_gep m 0 "m_mat" builder in ignore(L.build_store mat m_mat builder);
      let m_r = L.build_struct_gep m 1 "m_r" builder in ignore(L.build_store (L.const_int i32_t r) m_r builder);
      let m_c = L.build_struct_gep m 2 "m_c" builder in ignore(L.build_store (L.const_int i32_t c) m_c builder);
      m 
    in
    
    let build_default_mat (r,c) builder = 
      let mat = L.build_array_malloc float_t (L.build_mul r c "tmp" builder) "system_mat" builder in
      let m = L.build_alloca matrix_t "m" builder in
      let m_mat = L.build_struct_gep m 0 "m_mat" builder in ignore(L.build_store mat m_mat builder);
      let m_r = L.build_struct_gep m 1 "m_r" builder in ignore(L.build_store r m_r builder);
      let m_c = L.build_struct_gep m 2 "m_c" builder in ignore(L.build_store c m_c builder);
      m
    in
    let build_empty_mat builder = 
      let m = L.build_alloca matrix_t "m" builder in
      let m_r = L.build_struct_gep m 1 "m_r" builder in ignore(L.build_store (L.const_int i32_t 0) m_r builder);
      let m_c = L.build_struct_gep m 2 "m_c" builder in ignore(L.build_store (L.const_int i32_t 0) m_c builder);
      m
    in
    let build_matrix_elem_ptr s i j builder = 
      let mat = L.build_load (L.build_struct_gep s 0 "m_mat" builder) "mat" builder in
      let c = L.build_load (L.build_struct_gep s 2 "m_c" builder) "c_mat" builder in
      let index = L.build_add j (L.build_mul i c "tmp" builder) "index" builder in 
      L.build_gep mat [|index|] "element_ptr_ptr" builder 
    in

    let mat_float_mult m num builder =
      let s' = L.build_load (L.build_struct_gep m 0 "m_mat" builder) "m" builder in 
      let r = L.build_load (L.build_struct_gep m 1 "m_r" builder) "r_mat" builder in
      let c = L.build_load (L.build_struct_gep m 2 "m_c" builder) "c_mat" builder in
      let res_mat = build_default_mat (r,c) builder in 
      let res = L.build_load (L.build_struct_gep res_mat 0 "m_mat" builder) "mat" builder in
      ignore(L.build_call sconsmul_func [|s';res;r;c; num|] "" builder);
      res_mat
    in

    let mat_float_div m num builder =
      let s' = L.build_load (L.build_struct_gep m 0 "m_mat" builder) "m" builder in 
      let r = L.build_load (L.build_struct_gep m 1 "m_r" builder) "r_mat" builder in
      let c = L.build_load (L.build_struct_gep m 2 "m_c" builder) "c_mat" builder in
      let res_mat = build_default_mat (r,c) builder in 
      let res = L.build_load (L.build_struct_gep res_mat 0 "m_mat" builder) "mat" builder in
      ignore(L.build_call sconsdiv_func [|s';res;r;c; num|] "" builder);
      res_mat
    in
    (*elementwise matrix multiplication operator*)

    let mat_elemul_operation m1 m2 builder =
      let s1' = L.build_load (L.build_struct_gep m1 0 "m_mat" builder) "m" builder in 
      let s2' = L.build_load (L.build_struct_gep m2 0 "m_mat" builder) "m" builder in 
      let r1 = L.build_load (L.build_struct_gep m1 1 "m_r" builder) "r_mat" builder in
      let c1 = L.build_load (L.build_struct_gep m1 2 "m_c" builder) "c_mat" builder in
      let r2 = L.build_load (L.build_struct_gep m2 1 "m_r" builder) "r_mat" builder in
      let c2 = L.build_load (L.build_struct_gep m2 2 "m_c" builder) "c_mat" builder in
      
      let res_mat = build_default_mat (r1,c1) builder in 
      let res = L.build_load (L.build_struct_gep res_mat 0 "m_mat" builder) "mat" builder in
      ignore(L.build_call selemul_func [|s1'; s2'; res; r1; c1; r2; c2|] "" builder);
      res_mat
  in
    let mat_elediv_operation m1 m2 builder =
      let s1' = L.build_load (L.build_struct_gep m1 0 "m_mat" builder) "m" builder in 
      let s2' = L.build_load (L.build_struct_gep m2 0 "m_mat" builder) "m" builder in 
      let r1 = L.build_load (L.build_struct_gep m1 1 "m_r" builder) "r_mat" builder in
      let c1 = L.build_load (L.build_struct_gep m1 2 "m_c" builder) "c_mat" builder in
      let r2 = L.build_load (L.build_struct_gep m2 1 "m_r" builder) "r_mat" builder in
      let c2 = L.build_load (L.build_struct_gep m2 2 "m_c" builder) "c_mat" builder in
      
      let res_mat = build_default_mat (r1,c1) builder in 
      let res = L.build_load (L.build_struct_gep res_mat 0 "m_mat" builder) "mat" builder in
      ignore(L.build_call selediv_func [|s1'; s2'; res; r1; c1; r2; c2|] "" builder);
      res_mat
  in
    let mat_matmul_operation m1 m2 builder =
      let s1' = L.build_load (L.build_struct_gep m1 0 "m_mat" builder) "m" builder in 
      let s2' = L.build_load (L.build_struct_gep m2 0 "m_mat" builder) "m" builder in 
      let r1 = L.build_load (L.build_struct_gep m1 1 "m_r" builder) "r_mat" builder in
      let c1 = L.build_load (L.build_struct_gep m1 2 "m_c" builder) "c_mat" builder in
      let r2 = L.build_load (L.build_struct_gep m2 1 "m_r" builder) "r_mat" builder in
      let c2 = L.build_load (L.build_struct_gep m2 2 "m_c" builder) "c_mat" builder in
      let res_mat = build_default_mat (r1,c2) builder in 
      let res = L.build_load (L.build_struct_gep res_mat 0 "m_mat" builder) "mat" builder in
      ignore(L.build_call smatmul_func [|s1'; s2'; res; r1; c1; r2; c2|] "" builder);
      res_mat
  in
      let rec expr_ptr builder map se = 
        let e = snd se in match e with
          SId s -> lookup map s
        | SMatrixLit(f_array, (r,c)) -> build_matrix_lit (f_array, (r,c)) builder
        | SBinop (e1, op, ((A.Matrix,_ ) as e2))
        | SBinop ((A.Matrix,_ ) as e1, op, e2) -> (
          match (e1, op, e2) with
            ((A.Matrix,_), A.Div, (A.Float,_)) -> 
              let e1' = expr_ptr builder map e1 in 
              let e2' = expr builder map e2 in mat_float_div e1' e2' builder
          | ((A.Matrix,_), A.Mult, (A.Float,_)) -> 
              let e1' = expr_ptr builder map e1 in 
              let e2' = expr builder map e2 in mat_float_mult e1' e2' builder
          | ((A.Float,_), A.Mult, (A.Matrix,_)) -> 
          let e1' = expr builder map e1 in 
          let e2' = expr_ptr builder map e2 in mat_float_mult e2' e1' builder
          | ((A.Matrix,_), A.Mult, (A.Matrix,_)) -> 
              let e1' = expr_ptr builder map e1 in 
              let e2' = expr_ptr builder map e2 in mat_matmul_operation e1' e2' builder
          | ((A.Matrix,_), A.Elemult, (A.Matrix,_)) -> 
              let e1' = expr_ptr builder map e1 in 
              let e2' = expr_ptr builder map e2 in mat_elemul_operation e1' e2' builder
          | ((A.Matrix,_), A.Elediv, (A.Matrix,_)) -> 
              let e1' = expr_ptr builder map e1 in 
              let e2' = expr_ptr builder map e2 in mat_elediv_operation e1' e2' builder
          | _ -> raise (Failure "Not supported yet"))
        | SMatSlice (sem, e1, e2) ->
          let s' = expr_ptr builder map sem in
          let r = L.build_load (L.build_struct_gep s' 1 "m_r" builder) "r_mat" builder in
          let c = L.build_load (L.build_struct_gep s' 2 "m_c" builder) "c_mat" builder in
          let range_ind max = function
            SBeg -> L.const_int i32_t 0
          | SEnd -> L.build_sub max (L.const_int i32_t 1) ".tmp_sub" builder
          | SInd i -> expr builder map i
          in
          let range_inds max = function
            SRange(a,b) -> (range_ind max a),(range_ind max b)
          | _ -> let ind = expr builder map e1 in ind,ind
          in
          let i,j = range_inds r (snd e1) in
          let k,l = range_inds c (snd e2) in
          let res_mat = build_empty_mat builder in
          ignore(L.build_call smatslice_func [|s'; r; c; i; j; k; l; res_mat|] "" builder);
          res_mat
        | SCall ("trans", [e]) ->
          let s' = expr_ptr builder map e in
          let mat = L.build_load (L.build_struct_gep s' 0 "m_mat" builder) "mat" builder in
          let r = L.build_load (L.build_struct_gep s' 1 "m_r" builder) "r_mat" builder in 
          let c = L.build_load (L.build_struct_gep s' 2 "m_c" builder) "c_mat" builder in
          let res_mat = build_default_mat (c,r) builder in
          let res = L.build_load (L.build_struct_gep res_mat 0 "m_mat" builder) "mat" builder in
          ignore (L.build_call strans_func [|mat;res;r;c|] "" builder );
          res_mat

        | SCall ("empty", [r;c]) ->
          let r = expr builder map r in 
          let c = expr builder map c in
          let res_mat = build_default_mat (r,c) builder in
          res_mat
        | SCall ("imread", [f]) ->
          let f' = expr builder map f in
          let res_mat = build_empty_mat builder in 
          ignore(L.build_call simread_func [|f'; res_mat|] "" builder);
          res_mat
        | SCall (f, _) ->
            let (_, fdecl) = StringMap.find f function_decls in
            let tmp_var = L.build_alloca (ltype_of_typ fdecl.styp) ".ret" builder
            in ignore (L.build_store (expr builder map se) tmp_var builder);
            tmp_var
        | _ -> raise (Failure ("internal error: need alloc ptr for expression of type " ^ A.string_of_typ (fst se)))
      and expr builder map se = 
        let e = snd se in match e with
          SLiteral i -> L.const_int i32_t i
        | SBoolLit b  -> L.const_int i1_t (if b then 1 else 0)
        | SCharLit c  -> L.const_int i8_t (Char.code c)
        | SMatrixLit (_,_) -> L.build_load (expr_ptr builder map se) ".matlit" builder
        | SId s       -> L.build_load (expr_ptr builder map se) s builder
        | SStringLit s -> L.build_global_stringptr s ".strlit" builder
        | SFloatLit l -> L.const_float_of_string float_t l
        | SCastExpr (ct, ((et, _) as e)) ->
          let e' = expr builder map e in
          if et = ct then e' else let op = (match et,ct with
            A.Int,A.Float
          | A.Char,A.Float
          | A.Bool,A.Float -> L.build_sitofp
          | A.Float,A.Int
          | A.Float,A.Char
          | A.Float,A.Bool -> L.build_fptosi
          | A.Int,A.Char
          | A.Int,A.Bool
          | A.Char,A.Bool -> L.build_trunc
          | A.Bool, A.Char
          | A.Bool, A.Int
          | A.Char, A.Int -> L.build_sext
          | _,_ -> raise (Failure ("internal error: semant should have rejected cast " ^ A.string_of_typ et ^ " -> " ^ A.string_of_typ ct)))
          in op e' (ltype_of_typ ct) ".cast_op" builder
        | SNoexpr     -> L.const_int i32_t 0
        | SBinop ((A.Matrix,_ ), _, _)
        | SBinop (_, _, ((A.Matrix,_ ))) -> L.build_load (expr_ptr builder map se) ".matop" builder
        | SBinop ((A.Float,_ ) as e1, op, e2) ->
          let e1' =  expr builder map e1
          and e2' =  expr builder map e2
          in (match(e1, e2) with
          | _ -> (match op with 
              A.Add     -> L.build_fadd
            | A.Sub     -> L.build_fsub
            | A.Mult    -> L.build_fmul
            | A.Div     -> L.build_fdiv 
            | A.Equal   -> L.build_fcmp L.Fcmp.Oeq
            | A.Neq     -> L.build_fcmp L.Fcmp.One
            | A.Less    -> L.build_fcmp L.Fcmp.Olt
            | A.Leq     -> L.build_fcmp L.Fcmp.Ole
            | A.Greater -> L.build_fcmp L.Fcmp.Ogt
            | A.Geq     -> L.build_fcmp L.Fcmp.Oge
            | A.And | A.Or ->
              raise (Failure "internal error: semant should have rejected and/or on float")
            | A.Elemult | A.Elediv ->
              raise (Failure "internal error: semant should have rejected element mult/div on float")
          ) e1' e2' "tmp" builder)
        | SBinop (e1, op, e2) ->
          let e1' = expr builder map e1
          and e2' = expr builder map e2 in
          (match op with
            A.Add     -> L.build_add
          | A.Sub     -> L.build_sub
          | A.Mult    -> L.build_mul
          | A.Div     -> L.build_sdiv
          | A.And     -> L.build_and
          | A.Or      -> L.build_or
          | A.Equal   -> L.build_icmp L.Icmp.Eq
          | A.Neq     -> L.build_icmp L.Icmp.Ne
          | A.Less    -> L.build_icmp L.Icmp.Slt
          | A.Leq     -> L.build_icmp L.Icmp.Sle
          | A.Greater -> L.build_icmp L.Icmp.Sgt
          | A.Geq     -> L.build_icmp L.Icmp.Sge
          | A.Elemult | A.Elediv ->
            raise (Failure "internal error: semant should have rejected element mult/div on int")
          ) e1' e2' "tmp" builder


        | SMatSlice (_,_,_) -> L.build_load (expr_ptr builder map se) ".matslice" builder

        | SMatAccess (sem, e1, e2) -> 
            let idx = 
              let e1' = expr builder map e1 
              and e2' = expr builder map e2 
              and s' = expr_ptr builder map sem in
              build_matrix_elem_ptr s' e1' e2' builder
            in
            L.build_load idx "element_ptr" builder
              
        | SStringAccess (s, i) ->
            let idx =
              let s' = expr builder map s in
              let index = expr builder map i in
              L.build_gep s' [|index|] ".char_ptr" builder 
            in
            L.build_load idx ".char" builder
        | SUnop(op, ((t, _) as e)) ->
            let e' = expr builder map e in
          (match op with
            A.Neg when t = A.Float -> L.build_fneg 
          | A.Neg                  -> L.build_neg
          | A.Not                  -> L.build_not) e' "tmp" builder
        | SAssign ((_, s), e) -> let e' = expr builder map e in 
          let dst = (match s with
            SId v -> lookup map v
          | SMatAccess (sem, e1, e2) ->
              let e1' = expr builder map e1 
              and e2' = expr builder map e2 
              and s' = expr_ptr builder map sem in
              build_matrix_elem_ptr s' e1' e2' builder
          | SStringAccess (s, i) ->
              let s' = expr builder map s in
              let index = expr builder map i in
              L.build_gep s' [|index|] ".char_ptr" builder
          | _ -> raise (Failure "internal error: semant should have rejected assignment on invalid expression"))
          in ignore(L.build_store e' dst builder); e'
        | SCall ("height", [sem]) -> 
            let s' = expr_ptr builder map sem in
            L.build_load (L.build_struct_gep s' 1 "m_r" builder) "r_mat" builder
        | SCall ("width", [sem]) -> 
            let s' = expr_ptr builder map sem in
              L.build_load (L.build_struct_gep s' 2 "m_c" builder) "c_mat" builder
        | SCall ("printf", [e]) -> 
          L.build_call printf_func [| float_format_str ; (expr builder map e) |]
            "printf" builder
        | SCall ("printb", [e]) ->
          L.build_call printf_func [| int_format_str ; (expr builder map e) |]
            "printf" builder
        | SCall ("print", [e]) ->
          L.build_call printf_func [| int_format_str ; (expr builder map e) |]
            "printf" builder 
        | SCall ("prints", [e]) ->
          L.build_call printf_func [| string_format_str ; (expr builder map e) |]
            "printf" builder
        | SCall ("printc", [e]) ->
          L.build_call printf_func [| char_format_str ; (expr builder map e) |]
            "printf" builder
        | SCall ("printm", [sem]) ->
          let s' = expr_ptr builder map sem in 
          let mat = L.build_load (L.build_struct_gep s' 0 "m_mat" builder) "mat_mat" builder in 
          let r = L.build_load (L.build_struct_gep s' 1 "m_r" builder) "r_mat" builder in
          let c = L.build_load (L.build_struct_gep s' 2 "m_c" builder) "c_mat" builder in
          L.build_call sprintm_func [|mat; r; c|] "" builder
        | SCall ("sum", [sem]) -> 
          let s' = expr_ptr builder map sem in 
          let mat = L.build_load (L.build_struct_gep s' 0 "m_mat" builder) "mat_mat" builder in 
          let r = L.build_load (L.build_struct_gep s' 1 "m_r" builder) "r_mat" builder in
          let c = L.build_load (L.build_struct_gep s' 2 "m_c" builder) "c_mat" builder in
          L.build_call ssum_func [|mat; r; c|] "sum" builder
        | SCall ("mean", [sem]) -> 
          let s' = expr_ptr builder map sem in 
          let mat = L.build_load (L.build_struct_gep s' 0 "m_mat" builder) "mat_mat" builder in 
          let r = L.build_load (L.build_struct_gep s' 1 "m_r" builder) "r_mat" builder in
          let c = L.build_load (L.build_struct_gep s' 2 "m_c" builder) "c_mat" builder in
          L.build_call smean_func [|mat; r; c|] "mean" builder

        | SCall ("trans", _) -> L.build_load (expr_ptr builder map se) ".mattrans" builder
        | SCall ("empty", _) -> L.build_load (expr_ptr builder map se) ".matempty" builder

        | SCall ("filter2D", [src; dst; ddepth; kern]) ->
          let ss' = expr_ptr builder map src in
          let ds' = expr_ptr builder map dst in
          let dd = expr builder map ddepth in
          let ks' = expr_ptr builder map kern in
          L.build_call sfilter2D_func [|ss'; ds'; dd; ks'|] "" builder

        | SCall ("imread", _) -> L.build_load (expr_ptr builder map se) ".matret" builder

        | SCall ("imwrite", [f; m]) ->
          let f' = expr builder map f in
          let m' = expr_ptr builder map m in
          L.build_call simwrite_func [|f'; m'|] ".matret" builder

        | SCall ("invert", [src; dst;]) ->
          let ss' = expr_ptr builder map src in
          let ds' = expr_ptr builder map dst in
          L.build_call sinvert_func [|ss'; ds'|] "" builder
          
        | SCall (f, args) -> 
          
          let (fdef, fdecl) = StringMap.find f function_decls in
          let llargs = List.rev (List.map (expr builder map) (List.rev args)) in
          let result = (match fdecl.styp with 
                          A.Void -> ""
                        | _ -> f ^ "_result") in
          L.build_call fdef (Array.of_list llargs) result builder
        | SRange(_,_) -> raise (Failure "internal error: SRange should not reach codegen as expression")
        in

      (* Construct the function's "locals": formal arguments and locally
        declared variables.  Allocate each on the stack, initialize their
        value, if appropriate, and remember their values in the "locals" map *)
      let formal_vars = 
        let add_formal m (t, n) p = 
          L.set_value_name n p;
          let formal = L.build_alloca (ltype_of_typ t) n builder in
                ignore (L.build_store p formal builder);
          StringMap.add n formal m 
        (* Allocate space for any locally declared variables and add the
          * resulting registers to our map *)
        in List.fold_left2 add_formal StringMap.empty fdecl.sformals
          (Array.to_list (L.params the_function))

        in

        let add_terminal builder instr =
              match L.block_terminator (L.insertion_block builder) with
          Some _ -> ()
        | None -> ignore (instr builder) in

      let rec stmt (builder, vars, w_bb, wm_bb) = function
          SBlock sl ->
            let fold_stmt (builder, vars) = stmt (builder, vars, w_bb, wm_bb) in
            (fst (List.fold_left fold_stmt (builder, vars) sl), vars)
        | SExpr e -> ignore (expr builder vars e); (builder, vars)
        | SVarDecl vd -> (builder, match vd with
          | SDeclInit((t,n),e) -> 
              let e' = expr builder vars e 
              in L.set_value_name n e';
              let local_var = L.build_alloca (ltype_of_typ t) n builder
              in ignore (L.build_store e' local_var builder);
              StringMap.add n local_var vars
          | SDecl(t,n) -> 
              let local_var = L.build_alloca (ltype_of_typ t) n builder
              in StringMap.add n local_var vars)
        | SIf (predicate, then_stmt, else_stmt) ->
          let bool_val = expr builder vars predicate in
          let merge_bb = L.append_block context "merge" the_function in
                let build_br_merge = L.build_br merge_bb in (* partial function *)

          let then_bb = L.append_block context "then" the_function in
          add_terminal (fst (stmt ((L.builder_at_end context then_bb), vars, w_bb, wm_bb) then_stmt))
            build_br_merge;

          let else_bb = L.append_block context "else" the_function in
          add_terminal (fst (stmt ((L.builder_at_end context else_bb), vars, w_bb, wm_bb) else_stmt))
            build_br_merge;

          ignore(L.build_cond_br bool_val then_bb else_bb builder);
          (L.builder_at_end context merge_bb, vars)
        | SWhile (predicate, body) ->
          let pred_bb = L.append_block context "while" the_function in
          ignore(L.build_br pred_bb builder);

          let merge_bb = L.append_block context "merge" the_function in
          let body_bb = L.append_block context "while_body" the_function in
          add_terminal (fst (stmt ((L.builder_at_end context body_bb), vars, Some(body_bb), Some(merge_bb)) body))
            (L.build_br pred_bb);

          let pred_builder = L.builder_at_end context pred_bb in
          let bool_val = expr pred_builder vars predicate in

          ignore(L.build_cond_br bool_val body_bb merge_bb pred_builder);
          (L.builder_at_end context merge_bb, vars)

        (* Implement for loops as while loops *)
        | SFor (e1, e2, e3, body) -> stmt (builder, vars, w_bb, wm_bb)
          ( SBlock [SExpr e1 ; SWhile (e2, SBlock [body ; SExpr e3]) ] )

        | SBreak ->
            let while_merge_bb = (match wm_bb with
                Some(bb) -> bb
              | None -> raise (Failure "break statement not in while loop")
            ) in
            ignore(L.build_br while_merge_bb builder);
            (builder, vars)
        | SContinue ->
            let while_body_bb = (match w_bb with
                Some(bb) -> bb
              | None -> raise (Failure "break statement not in while loop")
            ) in
            ignore(L.build_br while_body_bb builder);
            (builder, vars)

        | SReturn e -> ignore(match fdecl.styp with
              (* Special "return nothing" instr *)
              A.Void -> L.build_ret_void builder 
              (* Build return statement *)
            | _ -> L.build_ret (expr builder vars e) builder ); (builder, vars)
      in
      let builder = fst(stmt (builder, formal_vars, None, None) (SBlock fdecl.sbody)) in

      add_terminal builder (match fdecl.styp with
          A.Void -> L.build_ret_void
        | t -> L.build_ret (L.const_int (ltype_of_typ t) 0))
    in 
    List.iter build_function_body functions;
  the_module
