/* Ocamlyacc parser for YAMML */

%{
open Ast
%}

%token SEMI COLON LPAREN RPAREN LBRACE RBRACE COMMA
%token LBRACK RBRACK
%token PLUS MINUS TIMES DIVIDE ASSIGN NOT
%token ELETIMES ELEDIVIDE
%token EQ NEQ LT LEQ GT GEQ TRUE FALSE AND OR
%token RETURN IF ELSE FOR WHILE CONTINUE BREAK
%token INT BOOL VOID CHAR STRING FLOAT MATRIX
%token HEIGHT WIDTH
%token <int> LITERAL
%token <char> CHARLIT
%token <string> STRINGLIT
%token <float> FLOATLIT
%token <string> ID
%token <string> IMPORT
%token EOF

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE
%left ELETIMES ELEDIVIDE
%right NOT NEG
%nonassoc LBRACK RBRACK
%nonassoc LPAREN RPAREN

%start program
%type <Ast.program> program

%%

program:
  imports decls EOF { match $2 with (v,f) -> ($1,v,f) }

imports:
   /* nothing */ { [] }
 | imports import { $2 :: $1 }

import:
  IMPORT { Import($1) }

decls:
   /* nothing */ { [], [] }
 | decls vdecl { ($2 :: fst $1), snd $1 }
 | decls fdecl { fst $1, ($2 :: snd $1) }

fdecl:
   typ ID LPAREN formals_opt RPAREN LBRACE stmt_list RBRACE
      { { typ = $1;
          fname = $2;
          formals = $4;
          body = List.rev $7 } }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { List.rev $1 }

formal_list:
    typ ID                   { [($1,$2)] }
  | formal_list COMMA typ ID { ($3,$4) :: $1 }

typ:
    INT { Int }
  | BOOL { Bool }
  | VOID { Void }
  | CHAR { Char }
  | STRING { String }
  | FLOAT { Float }
  | MATRIX { Matrix }

stmt_list:
    /* nothing */  { [] }
  | stmt_list stmt { $2 :: $1 }

stmt:
    expr_opt SEMI { Expr $1 }
  | RETURN expr_opt SEMI { Return $2 }
  | LBRACE stmt_list RBRACE { Block(List.rev $2) }
  | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
  | IF LPAREN expr RPAREN stmt ELSE stmt    { If($3, $5, $7) }
  | FOR LPAREN expr_opt SEMI expr SEMI expr_opt RPAREN stmt
     { For($3, $5, $7, $9) }
  | WHILE LPAREN expr RPAREN stmt { While($3, $5) }
  | CONTINUE SEMI { Continue }
  | BREAK SEMI { Break }
  | vdecl { VarDecl($1) }

vdecl:
   typ ID SEMI { Decl($1, $2) }
 | typ ID ASSIGN expr SEMI { DeclInit(($1, $2), $4) }

expr_opt:
    /* nothing */ { Noexpr }
  | expr          { $1 }

fst_ind_opt:
    /* nothing */ { Beg }
  | expr          { match $1 with a -> Ind(a) }

snd_ind_opt:
    /* nothing */ { End }
  | expr          { match $1 with a -> Ind(a) }

expr:
    LITERAL          { Literal($1) }
  | TRUE             { BoolLit(true) }
  | FALSE            { BoolLit(false) }
  | CHARLIT          { CharLit($1) }
  | STRINGLIT        { StringLit($1) }
  | FLOATLIT         { FloatLit(string_of_float $1) }
  | ID               { Id($1) }
  | mat_literal      { MatrixLit(fst $1, snd $1) }
  | expr PLUS   expr { Binop($1, Add,   $3) }
  | expr MINUS  expr { Binop($1, Sub,   $3) }
  | expr TIMES  expr { Binop($1, Mult,  $3) }
  | expr DIVIDE expr { Binop($1, Div,   $3) }
  | expr ELETIMES expr { Binop($1, Elemult, $3) }
  | expr ELEDIVIDE expr { Binop($1, Elediv, $3) }
  | expr EQ     expr { Binop($1, Equal, $3) }
  | expr NEQ    expr { Binop($1, Neq,   $3) }
  | expr LT     expr { Binop($1, Less,  $3) }
  | expr LEQ    expr { Binop($1, Leq,   $3) }
  | expr GT     expr { Binop($1, Greater, $3) }
  | expr GEQ    expr { Binop($1, Geq,   $3) }
  | expr AND    expr { Binop($1, And,   $3) }
  | expr OR     expr { Binop($1, Or,    $3) }
  | MINUS expr %prec NEG { Unop(Neg, $2) }
  | NOT expr         { Unop(Not, $2) }
  | expr ASSIGN expr   { Assign($1, $3) }
  | expr LBRACK expr COMMA expr RBRACK { 
    match $3, $5 with 
    | r, c -> MatAccess($1,r,c)}
  | expr LBRACK fst_ind_opt COLON snd_ind_opt COMMA fst_ind_opt COLON snd_ind_opt RBRACK { 
    match $3, $5, $7, $9 with 
    | a,b,c,d -> MatSlice($1,Range(a, b),Range(c, d))}
  | ID LPAREN actuals_opt RPAREN { Call($1, $3) }
  | LPAREN expr RPAREN { $2 }
  | LPAREN typ RPAREN expr { CastExpr($2, $4) }
  | expr LBRACK expr RBRACK { StringAccess($1, $3) }

mat_literal: 
    LBRACK RBRACK { [| |], (0, 0) }
  | LBRACK mat_assembly RBRACK { $2 } 

mat_assembly:
    ele { [| $1 |], (1, 1) }
  | mat_assembly COMMA ele { 
      let a,(r,c)  = (fst $1), (snd $1) in 
      Array.append a [| $3 |], (r, c+1) }
  | mat_assembly SEMI ele { 
      let a,(r,_)  = (fst $1), (snd $1) in 
      Array.append a [| $3 |], (r+1, 1) }

ele:
    FLOATLIT { $1 }
  | MINUS FLOATLIT %prec NEG { -. $2 }

actuals_opt:
    /* nothing */ { [] }
  | actuals_list  { List.rev $1 }

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }
