(*Semantically Checked AST *)

open Ast

type sexpr = typ * sx
and sx =
    SLiteral of int
  | SBoolLit of bool
  | SCharLit of char
  | SStringLit of string
  | SFloatLit of string
  | SMatrixLit of float array * (int * int)
  | SMatAccess of sexpr * sexpr * sexpr 
  | SMatSlice of sexpr * sexpr * sexpr
  | SStringAccess of sexpr * sexpr
  | SId of string
  | SBinop of sexpr * op * sexpr
  | SUnop of uop * sexpr
  | SAssign of sexpr * sexpr
  | SCall of string * sexpr list
  | SCastExpr of typ * sexpr
  | SNoexpr
  | SRange of sindex * sindex
  and sindex = SBeg | SEnd | SInd of sexpr

type sbind_decl = typ * string

type sbind_init = sbind_decl * sexpr

type sbind =
   SDecl of sbind_decl
 | SDeclInit of sbind_init

type sstmt =
    SBlock of sstmt list
  | SExpr of sexpr
  | SReturn of sexpr
  | SIf of sexpr * sstmt * sstmt
  | SFor of sexpr * sexpr * sexpr * sstmt
  | SWhile of sexpr * sstmt
  | SContinue
  | SBreak
  | SVarDecl of sbind

type sfunc_decl = {
    styp : typ;
    sfname : string;
    sformals : bind_decl list;
    sbody : sstmt list;
  }

type sprogram = bind list * sfunc_decl list

(* Pretty-printing functions *)


let string_of_sop = function
    Add -> "+"
  | Sub -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Equal -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq -> ">="
  | And -> "&&"
  | Or -> "||"
  | Elemult -> ".*"
  | Elediv -> "./"

let string_of_suop = function
    Neg -> "-"
  | Not -> "!"

let string_of_styp = function
    Int -> "int"
  | Bool -> "bool"
  | Void -> "void"
  | Char -> "char"
  | String -> "str"
  | Float -> "float"
  | Matrix -> "matrix"
  | Rangety -> "range"

let rec string_of_smatrix scalar cols _ =
  "[ " ^ (let rec string_of_elems = function
      [] -> ""
    | h::t -> string_of_sexpr h ^
              (if (List.length t) == 0 then ""
              else if (List.length t) mod cols == 0 then "; "
              else ", ") ^
              string_of_elems t
  in string_of_elems scalar) ^ " ]"

and string_of_sexpr (t,e) = 
  "(" ^ string_of_typ t ^ " : " ^ (match e with
    SLiteral(l) -> string_of_int l
  | SBoolLit(true) -> "true"
  | SBoolLit(false) -> "false"
  | SCharLit(c) -> "'" ^ String.make 1 c ^ "'"
  | SStringLit(s) -> "\"" ^ s ^ "\""
  | SFloatLit(f) -> f
  | SMatrixLit(s, (c, l)) -> let s_array = (Array.map (fun x -> Float, SFloatLit(Printf.sprintf "%g" x))) s in string_of_smatrix (Array.to_list s_array) c l
  | SMatAccess(m, c, r) -> string_of_sexpr m ^ "[" ^ string_of_sexpr c ^ "," ^ string_of_sexpr r ^ "]"
  | SMatSlice(m, r1, r2) -> string_of_sexpr m ^ "[" ^ string_of_sexpr r1 ^ "," ^ string_of_sexpr r2 ^ "]"
  | SStringAccess(s, i) -> string_of_sexpr s ^ "[" ^ string_of_sexpr i ^ "]"
  | SId(s) -> s
  | SBinop(e1, o, e2) ->
      string_of_sexpr e1 ^ " " ^ string_of_sop o ^ " " ^ string_of_sexpr e2
  | SUnop(o, e) -> string_of_suop o ^ string_of_sexpr e
  | SAssign(v, e) -> string_of_sexpr v ^ " = " ^ string_of_sexpr e
  | SCall(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_sexpr el) ^ ")"
  | SCastExpr(t, e) -> "(" ^ string_of_styp t ^ ") " ^ string_of_sexpr e
  | SNoexpr -> ""
  | SRange(s1, t1) -> (match s1, t1 with 
      SBeg, SEnd -> ":"
    | SBeg, SInd(i) -> ":"^string_of_sexpr i
    | SInd(i), SEnd -> string_of_sexpr i^":"
    | SInd(i1), SInd(i2) -> string_of_sexpr i1^":"^string_of_sexpr i2
    | _ -> raise (Failure "Semant should fail on invalid ranges")
  )
  )^ ")"

(* unsure if should change decl to sdecl*)
let string_of_svdecl = function
    SDecl(decl) -> string_of_styp (fst decl) ^ " " ^ (snd decl) ^ ";\n"
  | SDeclInit(decl, exp) -> string_of_styp (fst decl) ^ " " ^ (snd decl) ^ " = " ^ string_of_sexpr exp ^ ";\n"

let rec string_of_sstmt = function
    SBlock(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_sstmt stmts) ^ "}\n"
  | SExpr(expr) -> string_of_sexpr expr ^ ";\n";
  | SReturn(expr) -> "return " ^ string_of_sexpr expr ^ ";\n";
  | SIf(e, s, SBlock([])) -> "if (" ^ string_of_sexpr e ^ ")\n" ^ string_of_sstmt s
  | SIf(e, s1, s2) ->  "if (" ^ string_of_sexpr e ^ ")\n" ^
      string_of_sstmt s1 ^ "else\n" ^ string_of_sstmt s2
  | SFor(e1, e2, e3, s) ->
      "for (" ^ string_of_sexpr e1  ^ " ; " ^ string_of_sexpr e2 ^ " ; " ^
      string_of_sexpr e3  ^ ") " ^ string_of_sstmt s
  | SWhile(e, s) -> "while (" ^ string_of_sexpr e ^ ") " ^ string_of_sstmt s
  | SContinue -> "continue;\n"
  | SBreak -> "break;\n"
  | SVarDecl(v) -> string_of_svdecl v

let string_of_simport = function
  Import(filename) -> "#import <" ^ filename ^ ">\n"

let string_of_sfdecl fdecl =
  string_of_styp fdecl.styp ^ " " ^
  fdecl.sfname ^ "(" ^ String.concat ", " (List.map snd fdecl.sformals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_sstmt fdecl.sbody) ^
  "}\n"

let string_of_sprogram (vars, funcs) = String.concat "\n" (List.map string_of_svdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_sfdecl funcs)
